// Copyright (C) 2002 Andreas Stahel
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <iostream>
#include <fstream>
#include <math.h>
#include <string>

#include <octave/oct.h>
#include <octave/parse.h>

#define max(a,b) (((a)<(b)) ? (b) : (a))
#define min(a,b) (((a)<(b)) ? (a) : (b))

DEFUN_DLD (FEMEquation, args, , "[...] = FEMEquation (...)\n\
  sets up the system of linear equations for a numerical solution of a PDE\n\
\n\
  [A,b,n2d]=FEMEquation(nodes,elem,edges,'a','b','f','gD','gN')\n\
  [A,b,n2d]=FEMEquation(nodes,elem,edges,aVec,bVec,fVec,'gD','gN')\n\
     nodes elem edges  describe the mesh\n\
         see ReadMesh() for the description of the format\n\
     'a','b','f','gD','gN' are the names of the functions and coefficients\n\
         in the boundary value problem given below\n\
      the functions a, b and f may also be given as constant scalar value\n\
         or as vector with the values of the function at the nodes \n\
      the functions gD and gN may be given as constant scalar value\n\
\n\
  div(a*grad u) - b*u = f     in domain\n\
                    u = gD    on Dirichlet section of the boundary\n\
              a*du/dn = gN    on Neumann section of the boundary\n\
\n\
 A   is the matrix of the system to be solved.\n\
     It is stored in a symmetric, banded form (see SBSolve() )\n\
 b   is the RHS of the system to be solved\n\
 n2d is the renumbering of the nodes to the DOF of the system\n\
     n2d(k)=0  indicates that node k is a Dirichlet node\n\
     n2d(k)=nn indicates that the value of the solution at node k\n\
               is given by u(nn)\n\
\n\
see also ReadMesh, ShowMesh, FEMSolveSym, ShowSolution, FEMValue ")

{
  octave_value_list retval;
  int nargin = args.length ();
    if (nargin !=8 ) {
    print_usage ("FEMEquation");
    return retval;
  }
  using namespace std;
  octave_value_list argin;
  octave_value_list res;
  int nargout;

  Matrix nodes   = args(0).matrix_value();
  Matrix elem    = args(1).matrix_value();
  Matrix edges   = args(2).matrix_value();
  string Func;
  argin(0)=nodes;
  //  evaluate function a
  ColumnVector aV; 
  if (args(3).is_string()) {  // function given as string
    Func = args(3).string_value();
    nargout=1;
    res = feval (Func, argin, nargout);
    aV=res(0).column_vector_value();
  }
  else if(args(3).is_real_scalar()){ //function given as scalar
      aV.resize(nodes.rows());
      aV.fill(args(3).double_value());
  }
  else {  // function given by its values
    aV=args(3).column_vector_value();
  }

  // evaluate function b
  ColumnVector bV; 
  if (args(4).is_string()) {  // function given as string
    Func = args(4).string_value();
    nargout=1;
    res = feval (Func, argin, nargout);
    bV=res(0).column_vector_value();
  }
  else if(args(4).is_real_scalar()){ //function given as scalar
      bV.resize(nodes.rows());
      bV.fill(args(4).double_value());
  }
  else {  // function given by its values
    bV=args(4).column_vector_value();
  }

  // evaluate function f
  ColumnVector fV; 
  if (args(5).is_string()) {  // function given as string
    Func = args(5).string_value();
    nargout=1;
    res = feval (Func, argin, nargout);
    fV=res(0).column_vector_value();
  }
  else if(args(5).is_real_scalar()){ //function given as scalar
      fV.resize(nodes.rows());
      fV.fill(args(5).double_value());
  }
  else {  // function given by its values
    fV=args(5).column_vector_value();
  }

//  string gDFunc  = args(6).string_value();
  string gDFunc;
  double gDValue=0.0;;
  bool gDscalar=true;
  if (args(6).is_string()) {  // function given as string
    gDFunc = args(6).string_value();
    gDscalar=false;
  }
  else {gDValue=args(6).double_value();}

//  string gNFunc  = args(7).string_value();
  string gNFunc;
  double gNValue=0.0;
  bool gNscalar=true;
  if (args(7).is_string()) {  // function given as string
    gNFunc = args(7).string_value();
    gNscalar=false;
  }
  else {gNValue=args(7).double_value();}


  nargout=2;
  argin(0)=nodes;
  res = feval ("FindDOF", argin, nargout);
  int n  = res(0).int_value(); // actual number of DOF
  //ColumnVector n2d=res(1).column_vector_value();
  ColumnVector n2d=res(1).column_vector_value();
  

  // determine the bandwidth of the matrix
  int band=1;
  for(int k=0;k<elem.rows();k++){
    int n1,n2,n3;
    n1=(int)n2d((int)elem(k,0)-1);
    n2=(int)n2d((int)elem(k,1)-1);
    n3=(int)n2d((int)elem(k,2)-1);
    if (n1*n2>0) band = max(band,abs(n1-n2));
    if (n2*n3>0) band = max(band,abs(n2-n3));
    if (n3*n1>0) band = max(band,abs(n3-n1));
  }


  Matrix gMat(n,band+1);
  ColumnVector gVec(n);

  for(int k1=0;k1<n;k1++){
    gVec(k1)=0.0;
    for(int k2=0;k2<=band;k2++) gMat(k1,k2)=0.0;
  }


  Matrix corners(3,2);
  Matrix tmpCorners(1,2);
  double area;

  Matrix M(2,3);
  Matrix elMat (3,3);
  ColumnVector elVec (3);
  ColumnVector dofs(3);


  for(int k=0;k<elem.rows();k++){ // loop over all elements
    // read element information
    corners(0,0)=nodes((int)elem(k,0)-1,0); 
    corners(0,1)=nodes((int)elem(k,0)-1,1);
    corners(1,0)=nodes((int)elem(k,1)-1,0);
    corners(1,1)=nodes((int)elem(k,1)-1,1);
    corners(2,0)=nodes((int)elem(k,2)-1,0);
    corners(2,1)=nodes((int)elem(k,2)-1,1);

    area=((corners(1,0)-corners(0,0))*(corners(2,1)-corners(0,1))-
	  (corners(1,1)-corners(0,1))*(corners(2,0)-corners(0,0)))/2.0;

    M(0,0) = corners(2,1)-corners(1,1);
    M(0,1) = corners(0,1)-corners(2,1);
    M(0,2) = corners(1,1)-corners(0,1);
    M(1,0) = corners(2,0)-corners(1,0);
    M(1,1) = corners(0,0)-corners(2,0);
    M(1,2) = corners(1,0)-corners(0,0);

    double factor= 
	(aV((int)elem(k,0)-1)+aV((int)elem(k,1)-1)+aV((int)elem(k,2)-1))/
	(12.0*area);


    // factor* M'*M
    elMat(0,0)=factor*(M(0,0)*M(0,0)+M(1,0)*M(1,0)) + 
	bV((int)elem(k,0)-1)*area/3.0;
    elMat(0,1)=factor*(M(0,0)*M(0,1)+M(1,0)*M(1,1));
    elMat(0,2)=factor*(M(0,0)*M(0,2)+M(1,0)*M(1,2));

    //elMat(1,0)=factor*(M(0,1)*M(0,0)+M(1,1)*M(1,0));
    elMat(1,0)=elMat(0,1);
    elMat(1,1)=factor*(M(0,1)*M(0,1)+M(1,1)*M(1,1)) +
	bV((int)elem(k,1)-1)*area/3.0;
    elMat(1,2)=factor*(M(0,1)*M(0,2)+M(1,1)*M(1,2));
    
    //elMat(2,0)=factor*(M(0,2)*M(0,0)+M(1,2)*M(1,0));
    //elMat(2,1)=factor*(M(0,2)*M(0,1)+M(1,2)*M(1,1));
    elMat(2,0)=elMat(0,2);
    elMat(2,1)=elMat(1,2);
    elMat(2,2)=factor*(M(0,2)*M(0,2)+M(1,2)*M(1,2)) +
	bV((int)elem(k,2)-1)*area/3.0;


    //elVec(0)=fV(elem(k,0)-1)*area/3.0;
    //elVec(1)=fV(elem(k,1)-1)*area/3.0;
    //elVec(2)=fV(elem(k,2)-1)*area/3.0;
    // now element matrix and vector are generated
    // insert elements in  global matrix and vector
    dofs(0)=n2d((int)elem(k,0)-1);
    dofs(1)=n2d((int)elem(k,1)-1);
    dofs(2)=n2d((int)elem(k,2)-1);
    for (int k1=0;k1<3;k1++){ // loop over corners
      if (dofs(k1)>0){ // variable value at this corner
	for (int k2=0;k2<3;k2++){
	  if (dofs(k2)>0){ // variable value at this corner
	    if (k2>=k1){ 
	      gMat(min((int)dofs(k1),(int)dofs(k2))-1,
		   abs((int)(dofs(k1)-dofs(k2)))) 
		+= elMat(k1,k2);}
	  }
	  else {// Dirichlet node
	    if (gDscalar){
		gVec((int)dofs(k1)-1) += elMat(k1,k2)*gDValue;
	    }
	    else {
		Matrix tmpCorners(1,2);
		tmpCorners(0,0)=corners(k2,0);
		tmpCorners(0,1)=corners(k2,1);
		argin(0)=tmpCorners;
		nargout=1;
		res = feval (gDFunc, argin, nargout);
		gVec((int)dofs(k1)-1) += elMat(k1,k2)*res(0).double_value();
	    }
	  } // endif
	} // endfor k2
	gVec((int)dofs(k1)-1) += fV((int)elem(k,k1)-1)*area/3.0;
      } // endif dofs(k1)>0
    } //endfor k1
  } //endfor k (elements)


  // insert the edge contributions
  double factp=(1.0+1.0/sqrt(3.0))/2.0;
  double factn=(1.0-1.0/sqrt(3.0))/2.0;
  for(int k=0;k<edges.rows();k++){ // loop over all edges
    if (edges(k,2)>1){
      Matrix corners(2,2);
      corners(0,0)=nodes((int)edges(k,0)-1,0); 
      corners(0,1)=nodes((int)edges(k,0)-1,1);
      corners(1,0)=nodes((int)edges(k,1)-1,0); 
      corners(1,1)=nodes((int)edges(k,1)-1,1);
      Matrix points(2,2);
      points(0,0)=
       factp*corners(0,0)+factn*corners(1,0);
      points(0,1)=
       factp*corners(0,1)+factn*corners(1,1);
      points(1,0)=
       factn*corners(0,0)+factp*corners(1,0);
      points(1,1)=
       factn*corners(0,1)+factp*corners(1,1);
      argin(0)=points;
      ColumnVector resVec;
      if (gNscalar){
	  resVec.resize(2);
	  resVec.fill(gNValue);
      }
      else{
	  nargout=1;
	  res = feval(gNFunc, argin, nargout);
	  resVec=res(0).column_vector_value();
      }
      double w=(1.0-1.0/sqrt(3.0))/2.0;
      ColumnVector edgeVec(2);
      double length=
          sqrt( (corners(1,0)-corners(0,0))*(corners(1,0)-corners(0,0))+
                (corners(1,1)-corners(0,1))*(corners(1,1)-corners(0,1)) )/2.0;
      edgeVec(0) = ((1.0-w)*resVec(0)+w*resVec(1)) *length;
      edgeVec(1) = (w*resVec(0)+(1.0-w)*resVec(1)) *length;
      dofs(0)=n2d((int)edges(k,0)-1);dofs(1)=n2d((int)edges(k,1)-1);
      if (dofs(0)>0) { gVec((int)dofs(0)-1) -=  edgeVec(0);} 
      if (dofs(1)>0) { gVec((int)dofs(1)-1) -=  edgeVec(1);}
    }// endif
  }//endfor k


  retval(0)=gMat;
  retval(1)=gVec;
  retval(2)=n2d;


  return retval;

}



DEFUN_DLD (ReadMesh, args, , "[...] = ReadMesh (...)\n\
  reads the information of EasyMesh output into Octave data structures\n\
\n\
[nodes,elem,edges]=ReadMesh(filename);\n\
   the files filename.n filename.e filename.s are read\n\
   these files have to be generated first by EasyMesh filename\n\
   EasyMesh reads filename.d for the decription of the domain\n\
   The boundary markers in filename.d lead to the following \n\
     boundary conditions\n\
     1 leads to a Dirichlet condition\n\
     2 leads to a Neumann condition\n\
\n\
the matrices nodes elem edges  describe the mesh \n\
\n\
nodes contains the x and y coordinates of the nodes and the material\n\
   type at each node, i.e. nodes =[x1,y1,m1;x2,y2,m2;...;xn,yn,mn]\n\
elem contains the information about the elements\n\
   One row shows the numbers of the three nodes forming the element and\n\
   then the material type of the element\n\
edges contains the information about the boundary segments\n\
   One row shows the numbers of the two nodes forming the segment and\n\
   then the boundary markes is shown (1=Dirichlet, 2=Neumann)\n\
\n\
Currently material information is ignored\n\
\n\
see also ShowMesh, FEMEquation, FEMSolve, ShowSolution, FEMValue")

{
  octave_value_list retval;
  int nargin = args.length ();
    if (nargin !=1 ) {
    print_usage ("ReadMesh");
    return retval;
  }
  using namespace std;
  string basename  = args(0).string_value();
  string filename;

  int n ;
  int tmpNumb;
  char tmpChar;
  float tmpFloat;
  
  // reading nodes

  filename=basename;
  filename.append(".n");

  ifstream infile(filename.c_str(),ios::in);
  if(infile.fail()){cerr<<"Error: unable to open input file for nodes"<<endl;}

  infile>>n;
  Matrix nodes(n,3);
  for (int k=0;k<n;k++){
    infile>>tmpNumb;
    infile>>tmpChar;
    infile>>nodes(k,0);    infile>>nodes(k,1);    infile>>nodes(k,2);
  }
  infile.close();

  // reading elements

  filename=basename;
  filename.append(".e");

  infile.open(filename.c_str(),ios::in);
  if(infile.fail()){cerr<<"Error: unable to open input file for elements"<<endl;}

  infile>>n;
  Matrix elem(n,4);
  for (int k=0;k<n;k++){
    infile>>tmpNumb;
    infile>>tmpChar;
    infile>>elem(k,0);infile>>elem(k,1);infile>>elem(k,2);
    elem(k,0) +=1; elem(k,1) +=1; elem(k,2) +=1;
    for(int j=4;j<=9;j++) infile>>tmpNumb;
    for(int j=10;j<=11;j++) infile>>tmpFloat;
    infile>>elem(k,3);
  }
  infile.close();

  // reading edges 

  filename=basename;
  filename.append(".s");

  infile.open(filename.c_str(),ios::in);
  if(infile.fail()){cerr<<"Error: unable to open input file for edges"<<endl;}

  infile>>n;
  Matrix edges(n,3);
  for (int k=0;k<n;k++){
    infile>>tmpNumb;
    infile>>tmpChar;
    infile>>edges(k,0);infile>>edges(k,1);
    edges(k,0)+=1; edges(k,1)+=1;
    infile>>tmpNumb;infile>>tmpNumb;
    infile>>edges(k,2);
  }
  infile.close();

  int edgecount=0;
  for(int k=0;k<n;k++){
    if (edges(k,2)>0) edgecount++;
  }

  Matrix Nedges(edgecount,3);
  int ne=0;
  for(int k=0;k<n;k++){
    if (edges(k,2)>0) {
      Nedges(ne,0)=edges(k,0);
      Nedges(ne,1)=edges(k,1);
      Nedges(ne,2)=edges(k,2);
      ne++;
    }
  }


  retval(0)=nodes;
  retval(1)=elem;
  retval(2)=Nedges;
  return retval;

}


DEFUN_DLD (ReadMeshTriangle, args, , "[...] = ReadMeshTriangle (...)\n\
  reads the information of triangle output into Octave data structures\n\
\n\
[nodes,elem,edges]=ReadMeshTriangle(name)\n\
\n\
   the files name.node filename.ele filename.poly are read\n\
   if available  the file name.map (generated by CuthillMcKee) is read too\n\
   these files have to be generated first by triangle\n\
   triangle reads its inputfile for the decription of the domain\n\
   The boundary markers  lead to the following boundary conditions\n\
     1 leads to a Dirichlet condition\n\
     2 leads to a Neumann condition\n\
\n\
the matrices nodes, elem and edges  describe the mesh \n\
\n\
nodes contains the x and y coordinates of the nodes and the material \n\
   type at each node, i.e. nodes =[x1,y1,m1;x2,y2,m2;...;xn,yn,mn]\n\
elem contains the information about the elements\n\
   One row shows the numbers of the three nodes forming the element and\n\
   then the material type of the element\n\
edges contains the information about the boundary segments\n\
   One row shows the numbers of the two nodes forming the segment and\n\
   then the boundary markes is shown (1=Dirichlet, 2=Neumann)\n\
\n\
Currently material information is ignored\n\
\n\
see also ShowMesh, FEMEquation, FEMSolve, ShowSolution, FEMValue")

{
  octave_value_list retval;
  int nargin = args.length ();
    if (nargin !=1 ) {
    print_usage ("ReadMeshTriangle");
    return retval;
  }
  using namespace std;
  string basename  = args(0).string_value();
  string filename;
  
  int n=1 ;
  int tmpNumb;

  ColumnVector map(1);
  
  // reading the map file
  filename=basename;
  filename.append(".map");
  ifstream infile;
  bool mapflag=true;
  infile.open(filename.c_str(),ios::in);
  if(infile.fail()){mapflag=false;}
  else{
    infile>>n;
    map.resize(n);
    for (int k=0;k<n;k++){
      infile>>tmpNumb;
      infile>>map(k);
    }
//    infile.close();
  }
  infile.close();
  infile.clear();

  // reading nodes
  filename=basename;
  filename.append(".node");

  infile.open(filename.c_str(),ios::in);
  if(infile.fail())
     {cerr<<"Error: unable to open input file for nodes"<<endl;}
  infile>>n; 
  infile>>tmpNumb;infile>>tmpNumb;infile>>tmpNumb;
  if (mapflag==false){
    map.resize(n);
    for(int k=0;k<n;k++) map(k)=k+1;
  }
  Matrix nodes(n,3);
  for (int k=0;k<n;k++){
    infile>>tmpNumb;
    infile>>nodes((int)map(k)-1,0); 
    infile>>nodes((int)map(k)-1,1); 
    infile>>nodes((int)map(k)-1,2);
  }
  infile.close();
  infile.clear();

  // reading elements
  filename=basename;
  filename.append(".ele");

  infile.open(filename.c_str(),ios::in);
  if(infile.fail())
       {cerr<<"Error: unable to open input file for elements"<<endl;}
  infile>>n;
  Matrix elem(n,4);
  infile>> tmpNumb;  infile>> tmpNumb;
  for (int k=0;k<n;k++){
    infile>>tmpNumb;
    infile>>elem(k,0);infile>>elem(k,1);infile>>elem(k,2);
    elem(k,0)=map((int)elem(k,0)-1);
    elem(k,1)=map((int)elem(k,1)-1);
    elem(k,2)=map((int)elem(k,2)-1);
    elem(k,3)=0;
  }
  infile.close();
  infile.clear();

  // reading edges 
  filename=basename;
  filename.append(".poly");

  infile.open(filename.c_str(),ios::in);
  if(infile.fail())
      {cerr<<"Error: unable to open input file for edges"<<endl;}
  infile>>tmpNumb;  infile>>tmpNumb;  infile>>tmpNumb;  infile>>tmpNumb;
  infile>>n;  infile>>tmpNumb;
  Matrix edges(n,3);
  for (int k=0;k<n;k++){
    infile>>tmpNumb;
    infile>>edges(k,0);infile>>edges(k,1);infile>>edges(k,2);
    edges(k,0)=map((int)edges(k,0)-1); 
    edges(k,1)=map((int)edges(k,1)-1);
  }
  infile.close();
  infile.clear();

  retval(0)=nodes;
  retval(1)=elem;
  retval(2)=edges;
  return retval;

  }



DEFUN_DLD (WriteSolution, args, , "[...] = WriteSolution (...) bla ")
{
  octave_value_list retval;
  //int nargin = args.length ();
  //  if (nargin >2 ) {
  //  print_usage ("ElementContributionoct");
  //  return retval;
  //}
  using namespace std;
  Matrix nodes   = args(0).matrix_value();
  Matrix elem    = args(1).matrix_value();
  Matrix u = args(2).matrix_value();
  string filename=args(3).string_value();
//  string filename ="/tmp/meshdata.gnu";

  ofstream outfile(filename.c_str(),ios::out);
  if(outfile.fail()){cerr<<"Error: unable to open output file for solution"<<endl;}

  // coordinate axes
  // outfile<< "0.0 0.0 0.0\n";
  // outfile<< "0.0 0.0 1.0\n\n";
  // outfile<< "0.0 0.0 0.0\n";
  // outfile<< "0.0 1.0 0.0\n\n";
  // outfile<< "0.0 0.0 0.0\n";
  // outfile<< "1.0 0.0 0.0\n\n";

  // dummy line, otherwise gnuplot treats the file as a grid
  outfile<< "0.0 0.0 0.0\n\n";

  // write the solution
  for(int k=0;k<elem.rows();k++){
    outfile<<nodes((int)elem(k,0)-1,0)<<" "<<nodes((int)elem(k,0)-1,1)
          <<" "<<u((int)elem(k,0)-1,0)<<endl;
    outfile<<nodes((int)elem(k,1)-1,0)<<" "<<nodes((int)elem(k,1)-1,1)
          <<" "<<u((int)elem(k,1)-1,0)<<endl;
    outfile<<nodes((int)elem(k,2)-1,0)<<" "<<nodes((int)elem(k,2)-1,1)
          <<" "<<u((int)elem(k,2)-1,0)<<endl;
    outfile<<nodes((int)elem(k,0)-1,0)<<" "<<nodes((int)elem(k,0)-1,1)
	  <<" "<<u((int)elem(k,0)-1,0)<<endl<<endl;
  }

  outfile.close();
  return retval;
}


DEFUN_DLD (ShowSolutionMTV, args, , "[...] = ShowSolutionMTV (...)\n\
  generate the data to be ploted with plotmtv\n\
\n\
ShowSolutionMTV(nodes,elem,u,'filename')\n\
   nodes elem u  describe the mesh and the solution \n\
         see ReadMesh() for the description of the format\n\
  'filename' is the name of the file to be used\n\
\n\
see also ReadMesh, ShowMesh, FEMEquation, FEMSolve, ShowSolution, FEMValue ")
{
  octave_value_list retval;
  int nargin = args.length ();
    if (nargin !=4 ) {
    print_usage ("ShowSolutionMTV");
    return retval;
  }
  using namespace std;
  Matrix nodes   = args(0).matrix_value();
  Matrix elem    = args(1).matrix_value();
  Matrix u = args(2).matrix_value();
  string filename = args(3).string_value();

  //string filename ="./Solution.mtv";
  ofstream outfile(filename.c_str(),ios::out);
  if(outfile.fail()){cerr<<"Error: unable to open output file for solution"<<endl;}

  // Setup information
  outfile<< "$ DATA=CONTCURVE\n%  toplabel=\"u\"\n\n";
  outfile<< "% contstyle=1  # 1=contour lines, 2=colors, 3=mesh\n";
  outfile<< "% nsteps=10\n";
  outfile<< "#% contours=\"0.1 0.3 0.6\" # each contour value is specified\n";
  outfile<< "#% eyepos.x=0.5 eyepos.y=-1.5 eyepos.z=0.5 # another 3D view\n";
  outfile<< "% xlabel=\"x\"  ylabel=\"y\" zlabel=\"u\"\n\n";
  outfile<< "% leftworld=False   # right-handed coordinate system\n\n";

  // write the solution
  for(int k=0;k<elem.rows();k++){
    outfile<<nodes((int)elem(k,0)-1,0)<<" "<<nodes((int)elem(k,0)-1,1)
          <<" "<<u((int)elem(k,0)-1,0)<<endl;
    outfile<<nodes((int)elem(k,1)-1,0)<<" "<<nodes((int)elem(k,1)-1,1)
          <<" "<<u((int)elem(k,1)-1,0)<<endl;
    outfile<<nodes((int)elem(k,2)-1,0)<<" "<<nodes((int)elem(k,2)-1,1)
	   <<" "<<u((int)elem(k,2)-1,0)<<endl<<endl;
  }

  outfile.close();


  return retval;
}



DEFUN_DLD (FEMValue, args, nargout, "[...] = FEMValue (...)\n\
 ... = FEMValue(...)\n\
 evaluate the solution at a given set of points\n\
\n\
 values=FEMValue(xy,nodes,elem,u,defaultvalue)\n\
 values=FEMValue(xy,nodes,elem,u)\n\
\n\
 [values,grad]=FEMValue(xy,nodes,elem,u,defaultvalue)\n\
 [values,grad]=FEMValue(xy,nodes,elem,u)\n\
 \n\
 nodes elem u  information about the mesh and the solution\n\
 defaultvalue  value returned if a point is not in the domain\n\
               if not specified NaN is returned\n\
 xy  coordinates of the points at which the solution is to be evaluated\n\
\n\
 values  values of the solution at the points given by xy\n\
 grad    values of the gradient of the solution at the points given by xy")
{
  octave_value_list retval;
  int nargin = args.length ();
  if ((nargin<4)||(nargin>5) ) {
    print_usage ("FEMValue");
    return retval;
  }  

  double defaultvalue=0.0;
  bool gradient=false;
  if(nargout>=2) gradient=true;
  
  Matrix xy      = args(0).matrix_value();
  Matrix nodes   = args(1).matrix_value();
  Matrix elem    = args(2).matrix_value();
  ColumnVector u = args(3).column_vector_value();
  if (nargin==5) {defaultvalue = args(4).double_value();}

  int nx    = xy.rows();
  int nelem = elem.rows();

  Matrix MM(nelem,4);
  double dtmp;
  for(int k=0;k<nelem;k++){
    dtmp=min(nodes((int)elem(k,0)-1,0),nodes((int)elem(k,1)-1,0));
    MM(k,0)=min(dtmp,nodes((int)elem(k,2)-1,0));
    dtmp=max(nodes((int)elem(k,0)-1,0),nodes((int)elem(k,1)-1,0));
    MM(k,1)=max(dtmp,nodes((int)elem(k,2)-1,0));
    dtmp=min(nodes((int)elem(k,0)-1,1),nodes((int)elem(k,1)-1,1));
    MM(k,2)=min(dtmp,nodes((int)elem(k,2)-1,1));
    dtmp=max(nodes((int)elem(k,0)-1,1),nodes((int)elem(k,1)-1,1));
    MM(k,3)=max(dtmp,nodes((int)elem(k,2)-1,1));
  }

  double c1=0.0;
  double c2=0.0;
  ColumnVector values(nx);
  Matrix gradval(nx,2);
  for(int kxy=0;kxy<nx;kxy++){values(kxy)=defaultvalue;}

  bool flag=false;
  double eps=1.0e-12;

  double a11=0.0;  double a12=0.0;  double a21=0.0;  double a22=0.0;

  for(int kxy=0;kxy<nx;kxy++){
    int k=0;
    flag=false;  // set flag to true if correct element was found
    do{
    //for(int k=0;k<nelem;k++){//  # go through all elements
      if ((xy(kxy,0)>=MM(k,0)-eps)&&(xy(kxy,0)<=MM(k,1)+eps)&&
          (xy(kxy,1)>=MM(k,2)-eps)&&(xy(kxy,1)<=MM(k,3)+eps)){
	// examine this element closely
        a11=nodes((int)elem(k,1)-1,0)-nodes((int)elem(k,0)-1,0);
        a12=nodes((int)elem(k,1)-1,1)-nodes((int)elem(k,0)-1,1);
        a21=nodes((int)elem(k,2)-1,0)-nodes((int)elem(k,0)-1,0);
        a22=nodes((int)elem(k,2)-1,1)-nodes((int)elem(k,0)-1,1);

        double dx1=xy(kxy,0)-nodes((int)elem(k,0)-1,0);
        double dx2=xy(kxy,1)-nodes((int)elem(k,0)-1,1);

	// solve the system
	double det=a11*a22-a21*a12;
	c1=( a22*dx1-a21*dx2)/det;
	c2=(-a12*dx1+a11*dx2)/det;
	if ((c1>=-eps)&&(c2>=-eps)&&((c1+c2)<=(1+eps)))
	  flag=true;
      }//endif # examine this element closely
      //} //endfor # k
      k++;
    }while((flag==false)&&(k<=nelem));
    if(flag){
      k--;
      double u0=u((int)elem(k,0)-1);
      double u1=u((int)elem(k,1)-1);
      double u2=u((int)elem(k,2)-1);
      values(kxy)=u0 + c1*(u1-u0) + c2*(u2-u0);
      if (gradient) {
	double area= a11*a22-a21*a12;
	gradval(kxy,0) = (-a12*(u2-u0) + (u1-u0)*a22)/area;
	gradval(kxy,1) = (+a11*(u2-u0) - (u1-u0)*a21)/area;
      }
    }
    else {
      values(kxy)=defaultvalue;
      if (gradient) {gradval(kxy,0)=gradval(kxy,1)=defaultvalue;}
    }
  } //endfor # kxy


  retval(0)=values;
  if (gradient) retval(1) = gradval;
  
  return retval;
}

DEFUN_DLD (FEMIntegrate, args, , "[...] = FEMIntegrate (...)\n\
  integrate a function over the mesh\n\
\n\
  integral = FEMIntegrate(nodes,elem,'f')\n\
  integral = FEMIntegrate(nodes,elem,fValues)\n\
     nodes elem  describe the mesh \n\
                 see ReadMesh() for the description of the format\n\
     'f'         is the name of the function to be integrated\n\
     fValues     a vector with the values of the function at the nodes\n\
\n\
     integral    is the integral of the given function over the mesh")

{
  octave_value_list retval;
  int nargin = args.length ();
  if (nargin !=3 ) {
    print_usage ("FEMIntegrate");
    return retval;
  }
  using namespace std;
  octave_value_list argin;
  octave_value_list res;
  int nargout;
  
  Matrix nodes   = args(0).matrix_value();
  Matrix elem    = args(1).matrix_value();
  // read the function values
  string fFunc;
  ColumnVector fV; 
  if (args(2).is_string()) {  // function given as string
    fFunc = args(2).string_value();
    nargout=1;
    argin(0)=nodes;
    res = feval (fFunc, argin, nargout);
    fV=res(0).column_vector_value();
  }
  else if(args(2).is_real_scalar()){ //function given as scalar
      fV.resize(nodes.rows());
      fV.fill(args(2).double_value());
  }
  else {  // function given by its values
    fV=args(2).column_vector_value();
  }

  Matrix corners(3,2);
  double areaThird;
  double integral=0.0;

  for(int k=0;k<elem.rows();k++){ // loop over all elements
    // read element information
    corners(0,0)=nodes((int)elem(k,0)-1,0);
    corners(0,1)=nodes((int)elem(k,0)-1,1);
    corners(1,0)=nodes((int)elem(k,1)-1,0);
    corners(1,1)=nodes((int)elem(k,1)-1,1);
    corners(2,0)=nodes((int)elem(k,2)-1,0);
    corners(2,1)=nodes((int)elem(k,2)-1,1);

    areaThird=fabs(((corners(1,0)-corners(0,0))*(corners(2,1)-corners(0,1))-
	  (corners(1,1)-corners(0,1))*(corners(2,0)-corners(0,0))))/2.0/3.0;

    integral +=(fV((int)elem(k,0)-1)+fV((int)elem(k,1)-1)
		+fV((int)elem(k,2)-1))*areaThird;
  }
  retval(0)=integral;
  return retval;
}

DEFUN_DLD (FEMGradient, args, , "[...] = FEMGradient (...)\n\
  determine the values of the gradient at the nodes\n\
\n\
  grad = FEMGradient(nodes,elem,u)\n\
  integral = FEMIntegrate(nodes,elem,fValues)\n\
     nodes elem  describe the mesh \n\
                 see ReadMesh() for the description of the format\n\
     u           are the values at the nodes\n\
\n\
     grad        are the components of the gradient at the nodes")

{
  octave_value_list retval;
  int nargin = args.length ();
  if (nargin !=3 ) {
    print_usage ("FEMGradient");
    return retval;
  }
  
  Matrix nodes   = args(0).matrix_value();
  Matrix elem    = args(1).matrix_value();
  ColumnVector u = args(2).column_vector_value();

  Matrix corners(3,2);
  Matrix M(2,3);
  Matrix grad(nodes.rows(),2);
  grad.fill(0.0);
  ColumnVector weights(nodes.rows());
  weights.fill(0.0);
  ColumnVector elementGrad(2);
  double area2=0.0;
  double angle=0.0;

  for(int k=0;k<elem.rows();k++){ // loop over all elements
    // read element information
    corners(0,0)=nodes((int)elem(k,0)-1,0);    
    corners(0,1)=nodes((int)elem(k,0)-1,1);
    corners(1,0)=nodes((int)elem(k,1)-1,0);
    corners(1,1)=nodes((int)elem(k,1)-1,1);
    corners(2,0)=nodes((int)elem(k,2)-1,0);
    corners(2,1)=nodes((int)elem(k,2)-1,1);

    M(0,0) = corners(2,1)-corners(1,1);
    M(0,1) = corners(0,1)-corners(2,1);
    M(0,2) = corners(1,1)-corners(0,1);
    M(1,0) = corners(2,0)-corners(1,0);
    M(1,1) = corners(0,0)-corners(2,0);
    M(1,2) = corners(1,0)-corners(0,0);

    area2=(M(1,2)*M(0,1)-M(0,2)*M(1,1)); // the double of the area
    
    // compute the component of the gradient on the element
    elementGrad(0)= (M(0,0)*u((int)elem(k,0)-1)+M(0,1)*u((int)elem(k,1)-1)
                     +M(0,2)*u((int)elem(k,2)-1))/area2;
    elementGrad(1)=-(M(1,0)*u((int)elem(k,0)-1)+M(1,1)*u((int)elem(k,1)-1)
                     +M(1,2)*u((int)elem(k,2)-1))/area2;

    // first angle with help of scalar product
    angle=acos((-M(1,2)*M(1,1)-M(0,2)*M(0,1))/
	       ( sqrt(M(1,2)*M(1,2)+M(0,2)*M(0,2))
                *sqrt(M(1,1)*M(1,1)+M(0,1)*M(0,1))));
    grad((int)elem(k,0)-1,0)+=elementGrad(0)*angle;
    grad((int)elem(k,0)-1,1)+=elementGrad(1)*angle;
    weights((int)elem(k,0)-1) += angle;

    // second angle with help of scalar product
    angle=acos((-M(1,0)*M(1,2)-M(0,0)*M(0,2))/
	       ( sqrt(M(0,0)*M(0,0)+M(1,0)*M(1,0))
                *sqrt(M(0,2)*M(0,2)+M(1,2)*M(1,2))));
    grad((int)elem(k,1)-1,0)+=elementGrad(0)*angle;
    grad((int)elem(k,1)-1,1)+=elementGrad(1)*angle;
    weights((int)elem(k,1)-1) += angle;

    // third angle with help of scalar product
    angle=acos((-M(1,0)*M(1,1)-M(0,0)*M(0,1))/
	       ( sqrt(M(1,0)*M(1,0)+M(0,0)*M(0,0))
                *sqrt(M(1,1)*M(1,1)+M(0,1)*M(0,1))));
    grad((int)elem(k,2)-1,0)+=elementGrad(0)*angle;
    grad((int)elem(k,2)-1,1)+=elementGrad(1)*angle;
    weights((int)elem(k,2)-1) += angle;
  }
 
  for(int k=0;k<nodes.rows();k++){ // loop over all nodes
    grad(k,0) /= weights(k);
    grad(k,1) /= weights(k);
  }

  retval(0)=grad;
  return retval;
}
